import socket
import ssl

from .packet import Packet

import bson


def getBookingData():
    hostname = "booking-loco.kakao.com"
    context = ssl.create_default_context()

    with socket.create_connection((hostname, 443)) as sock:
        with context.wrap_socket(sock, server_hostname=hostname) as ssock:
            data = bson.encode({"os": "win32", "model": "", "MCCMNC": ""})

            b = Packet(1000, 0, "GETCONF", 0, data)
            ssock.write(b.toLocoPacket())

            data = ssock.recv(4096)

            recvPacket = Packet()
            recvPacket.readLocoPacket(data)
            return recvPacket



import os
import io
import struct


try:
    from Cryptodome.Cipher import AES, PKCS1_OAEP
    from Cryptodome.Hash import SHA1
    from Cryptodome.PublicKey.RSA import construct
    from Cryptodome.Signature import pss
except ImportError as e:
    print("ImportError: {}".format(str(e)))
    exit()


class CryptoManager:
    def __init__(self):
        self.aes_key = os.urandom(16)

    def getRsaPublicKey(self):
        n = int(
            "A44960441C7E83BB27898156ECB13C8AFAF05D284A4D1155F255CD22D3176CDE50482F2F27F71348E4D2EB5F57BF9671EF15C9224E042B1B567AC1066E06691143F6C50F88787F68CF42716B210CBEF0F59D53405A0A56138A6872212802BB0AEEA6376305DBD428831E8F61A232EFEDD8DBA377305EF972321E1352B5F64630993E5549C64FCB563CDC97DA2124B925DDEA12ADFD00138910F66937FAB68486AE43BFE203C4A617F9F232B5458A9AB409BAC8EDADEF685545F9B013986747737B3FD76A9BAC121516226981EA67225577D15D0F082B8207EAF7CDCB13123937CB12145837648C2F3A65018162315E77EAD2D2DD5986E46251764A43B9BA8F79",
            16,
        )
        e = int("3", 16)

        rsa_key = construct((n, e))
        return rsa_key

    def getHandshakePacket(self):
        f = io.BytesIO()

        enced = self.rsaEncrypt(self.aes_key)
        f.write(struct.pack("<I", len(enced)))
        f.write(struct.pack("<I", 12))
        f.write(struct.pack("<I", 2))
        f.write(enced)

        return f.getvalue()

    def rsaEncrypt(self, data):
        rsa_key = self.getRsaPublicKey()
        rsa_chiper = PKCS1_OAEP.new(
            key=rsa_key, hashAlgo=SHA1, mgfunc=lambda x, y: pss.MGF1(x, y, SHA1)
        )

        return rsa_chiper.encrypt(data)

    def aesEncrypt(self, data, iv):
        aes_chiper = AES.new(self.aes_key, AES.MODE_CFB, iv, segment_size=128)
        return aes_chiper.encrypt(data)

    def aesDecrypt(self, data, iv):
        aes_chiper = AES.new(self.aes_key, AES.MODE_CFB, iv, segment_size=128)
        return aes_chiper.decrypt(data)



import json

from .httpApi import upload

import hashlib
import requests


class Message:
    def __init__(self, channel, body):
        self.channel = channel
        self.rawBody = body
        self.logId = self.rawBody["chatLog"]["logId"]
        self.type = self.rawBody["chatLog"]["type"]
        self.message = self.rawBody["chatLog"]["message"]
        self.id = self.rawBody["chatLog"]["msgId"]
        self.author = self.rawBody["chatLog"]["authorId"]

        try:
            if "attachment" in self.rawBody["chatLog"]:
                self.attachment = json.loads(self.rawBody["chatLog"]["attachment"])
            else:
                self.attachment = {}
        except:
            pass

        self.nickName = self.author

    def __repr__(self):
        return "<Message id={0.id} channel={0.channel!r} type={0.type!r} author={0.author!r}>".format(
            self
        )

    async def reply(self, msg, t=1):
        return await self.channel.sendChat(
            msg,
            json.dumps(
                {
                    "attach_only": False,
                    "attach_type": t,
                    "mentions": [],
                    "src_linkId": self.channel.li,
                    "src_logId": self.logId,
                    "src_mentions": [],
                    "src_message": self.message,
                    "src_type": self.type,
                    "src_userId": self.author,
                }
            ),
            26,
        )

    async def sendChat(self, msg, extra, t):
        return await self.channel.sendChat(msg, extra, t)

    async def sendText(self, msg):
        return await self.channel.sendText(msg)

    async def delete(self):
        return await self.channel.deleteMessage(self.logId)

    async def hide(self):
        return await self.channel.hideMessage(self.logId, self.type)

    async def kick(self):
        return await self.channel.kickMember(self.authorId)

    async def sendPhoto(self, data, w, h):
        path, key, url = upload(data, "image/jpeg", self.authorId)
        return await self.channel.forwardChat(
            "",
            json.dumps(
                {
                    "thumbnailUrl": url,
                    "thumbnailHeight": w,
                    "thumbnailWidth": h,
                    "url": url,
                    "k": key,
                    "cs": hashlib.sha1(data).hexdigest().upper(),
                    "s": len(data),
                    "w": w,
                    "h": h,
                    "mt": "image/jpeg",
                }
            ),
            2,
        )

    async def sendPhotoPath(self, path, w, h):
        with open(path, "rb") as f:
            data = f.read()

        return await self.sendPhoto(data, w, h)

    async def sendPhotoUrl(self, url, w, h):
        r = requests.get(url)
        r.raise_for_status()

        return await self.sendPhoto(r.content, w, h)

    async def sendLongText(self, title, content):
        path, key, url = upload(content.encode("utf-8"), "image/jpeg", self.authorId)

        return await self.channel.forwardChat(
            title,
            json.dumps(
                {"path": path, "k": key, "s": len(content), "cs": "", "sd": True}
            ),
            1,
        )



from socket import socket

from .cryptoManager import CryptoManager
from .packet import Packet

import bson


def getCheckInData(host: str, port: int):
    crypto = CryptoManager()

    sock = socket()
    sock.connect((host, port))

    handshakePacket = crypto.getHandshakePacket()
    sock.send(handshakePacket)

    p = Packet(
        1,
        0,
        "CHECKIN",
        0,
        bson.encode(
            {
                "userId": 0,
                "os": "win32",
                "ntype": 0,
                "appVer": "3.14",
                "MCCMNC": "999",
                "lang": "ko",
            }
        ),
    )

    sock.send(p.toEncryptedLocoPacket(crypto))

    data = sock.recv(2048)

    recvPacket = Packet()
    recvPacket.readEncryptedLocoPacket(data, crypto)

    return recvPacket



import time

from .packet import Packet

import bson


class Channel:
    def __init__(self, chatId, li, writer):
        self.chatId = chatId
        self.li = li
        self.writer = writer

    async def __sendPacket(self, command, data):
        packet = Packet(0, 0, command, 0, bson.encode(data))
        return (await self.writer.sendPacket(packet)).toJsonBody()

    async def sendChat(self, msg, extra, t):
        data = {
            "chatId": self.chatId,
            "extra": extra,
            "type": t,
            "msgId": time.time(),
            "msg": str(msg),
            "noSeen": False,
        }
        return await self.__sendPacket("WRITE", data)

    async def send(self, msg):
        return await self.sendChat(msg, "{}", 1)

    async def forwardChat(self, msg, extra, t):
        data = {
            "chatId": self.chatId,
            "extra": extra,
            "type": t,
            "msgId": time.time(),
            "msg": str(msg),
            "noSeen": False,
        }
        return await self.__sendPacket("FORWARD", data)

    async def deleteMessage(self, logId):
        data = {"chatId": self.chatId, "logId": logId}
        return await self.__sendPacket("DELETEMSG", data)

    async def hideMessage(self, logId, t):
        if self.li:
            data = {"c": self.chatId, "li": self.li, "logId": logId, "t": t}
            return await self.__sendPacket("REWRITE", data)

    async def kickMember(self, mid):
        if self.li:
            data = {
                "li": self.li,
                "c": self.chatId,
                "mid": mid,
            }
            await self.__sendPacket("KICKMEM", data)

    async def setMeta(self, t, content):
        data = {"chatId": self.chatId, "type": t, "content": content}
        return await self.__sendPacket("SETMETA", data)

    async def getLinkInfo(self):
        return await self.__sendPacket("INFOLINK", {"lis": [self.li]})

    async def getChatInfo(self):
        return await self.__sendPacket("CHATINFO", {"chatId": self.chatId})

    async def getUserInfo(self, userId):
        return await self.__sendPacket(
            "MEMBER", {"chatId": self.chatId, "memberIds": [userId]}
        )



"""
kakao.py
=======
kakao.py is a very simple kakaotalk LOCO/HTTP API protocol wrapper for python.

-------------

Introduction
------------
Loco protocol compatible python library

This is rewritten version of KakaoBot by jhlee.

DO NOT USE IN ABUSING. ABUSING CAN CAUSE PERMANENT SERVICE RESTRICTION.

Quick Start
-------
```python
import kakao

class Myclass(kakao.Client):
    async def onReady(self):
        print("Logged on")

    async def onMessage(self, chat):
        if chat.message == "ping":
            await chat.reply("pong!")

kakao.check_reg("LoginId", "LoginPw")
client = Myclass("LoginId", "LoginPw")
client.run()
```

Thanks to
-------
idev-develop/KakaoBot

License
-------
MIT Licence
"""

import json
from socket import socket
import asyncio
import struct

from .booking import getBookingData
from .checkIn import getCheckInData
from .cryptoManager import CryptoManager
from .httpApi import postText, Login
from .writer import Writer
from .chat import Message
from .channel import Channel
from .packet import Packet

try:
    import bson
except ImportError as e:
    print("ImportError: {}".format(str(e)))
    exit()


class Client:
    """
    Represents a client connection that connects to KakaoTalk.

    Parameters:
        LoginId: account ID [type str]
        LoginPw: account PW [type str]
        device_name : Device's Name. [type str]
        device_uuid : Device's Uuid. [type str]
        debug: if debug is true, it prints some log to console. [type bool]

    Returns:

    Remarks:
        You can pass any string to 'device_uuid' but base64-encoded string recommended.
        Default value of device_name is "kakao.py".
        Default value of device_uuid id base64-encoded str of "kakaopydev1202".
    """

    def __init__(
        self,
        LoginId,
        LoginPw,
        device_name="kakao.py",
        device_uuid="a2FrYW9weWRldjEyMDI=",
        debug=False,
    ):
        self.__sock: socket
        self.__StreamReader: asyncio.StreamReader
        self.__StreamWriter: asyncio.StreamWriter
        self.__crypto: CryptoManager
        self.__writer: Writer
        self.__accessKey: str

        self.LoginId = LoginId
        self.LoginPw = LoginPw
        self.device_name = device_name
        self.device_uuid = device_uuid
        self.debug = debug

        self.__packetID = 0

        self.__processingBuffer = b""
        self.__processingHeader = b""
        self.__processingSize = 0

        self.packetDict = {}

    def postText(self, chatId, li, text, notice=False):
        """
        Description.

        Parameters:

        Returns:

        Remarks:
        """
        postText(chatId, li, text, notice, self.__accessKey, self.device_uuid)

    async def __recvPacket(self):
        encryptedBuffer = b""
        currentPacketSize = 0

        while True:
            recv = await self.__StreamReader.read(256)

            if not recv:
                print(recv)
                self.loop.stop()
                break

            encryptedBuffer += recv

            if not currentPacketSize and len(encryptedBuffer) >= 4:
                currentPacketSize = struct.unpack("<I", encryptedBuffer[0:4])[0]

            if currentPacketSize:
                encryptedPacketSize = currentPacketSize + 4

                if len(encryptedBuffer) >= encryptedPacketSize:
                    self.loop.create_task(
                        self.__processingPacket(encryptedBuffer[0:encryptedPacketSize])
                    )
                    encryptedBuffer = encryptedBuffer[encryptedPacketSize:]
                    currentPacketSize = 0

    async def __processingPacket(self, encryptedPacket):
        encLen = encryptedPacket[0:4]
        IV = encryptedPacket[4:20]
        BODY = encryptedPacket[20:]

        self.__processingBuffer += self.__crypto.aesDecrypt(BODY, IV)

        if not self.__processingHeader and len(self.__processingBuffer) >= 22:
            self.__processingHeader = self.__processingBuffer[0:22]
            self.__processingSize = (
                struct.unpack("<i", self.__processingHeader[18:22])[0] + 22
            )

        if self.__processingHeader:
            if len(self.__processingBuffer) >= self.__processingSize:
                p = Packet()
                p.readLocoPacket(self.__processingBuffer[: self.__processingSize])

                self.loop.create_task(self.__onPacket(p))

                self.__processingBuffer = self.__processingBuffer[
                    self.__processingSize :
                ]
                self.__processingHeader = b""

    async def __onPacket(self, packet):
        if packet.PacketID in self.packetDict:
            self.packetDict[packet.PacketID].set_result(packet)
            del self.packetDict[packet.PacketID]

        self.loop.create_task(self.onPacket(packet))

        body = packet.toJsonBody()

        if self.debug == True:
            print(packet.PacketName)

        if packet.PacketName == "MSG":
            chatId = body["chatLog"]["chatId"]

            if "li" in body:
                li = body["li"]
            else:
                li = 0

            channel = Channel(chatId, li, self.__writer)
            chat = Message(channel, body)

            self.loop.create_task(self.onMessage(chat))

        if packet.PacketName == "NEWMEM":
            chatId = body["chatLog"]["chatId"]

            if "li" in body:
                li = body["li"]
            else:
                li = 0

            channel = Channel(chatId, li, self.__writer)
            self.loop.create_task(self.onJoin(packet, channel))

        if packet.PacketName == "DELMEM":
            chatId = body["chatLog"]["chatId"]

            if "li" in body:
                li = body["li"]
            else:
                li = 0

            channel = Channel(chatId, li, self.__writer)
            self.loop.create_task(self.onQuit(packet, channel))

        if packet.PacketName == "DECUNREAD":
            chatId = body["chatId"]

            channel = Channel(chatId, 0, self.__writer)
            self.loop.create_task(self.onRead(channel, body))

    async def onReady(self):
        pass

    async def onPacket(self, packet):
        pass

    async def onMessage(self, chat):
        pass

    async def onJoin(self, packet, channel):
        pass

    async def onQuit(self, packet, channel):
        pass

    async def onRead(self, channel, packet):
        pass

    async def __heartbeat(self):
        while True:
            await asyncio.sleep(180)
            PingPacket = Packet(0, 0, "PING", 0, bson.encode({}))
            self.loop.create_task(self.__writer.sendPacket(PingPacket))

    async def __login(
        self,
        LoginId,
        LoginPw,
    ):
        r = json.loads(Login(LoginId, LoginPw, self.device_name, self.device_uuid))

        if r["status"] == -101:
            print("이전에 로그인이 되어있는 PC에서 로그아웃 해주세요")

        elif r["status"] == -100:
            print("디바이스 등록이 되어 있지 않습니다")

        elif r["status"] == 12:
            print("카카오계정 또는 비밀번호를 다시 확인해 주세요")

        if r["status"] != 0:
            self.loop.stop()
            raise Exception(str(r))

        self.__accessKey = r["access_token"]

        bookingData = getBookingData().toJsonBody()

        checkInData = getCheckInData(
            bookingData["ticket"]["lsl"][0], bookingData["wifi"]["ports"][0]
        ).toJsonBody()

        self.__StreamReader, self.__StreamWriter = await asyncio.open_connection(
            checkInData["host"], int(checkInData["port"])
        )

        self.__crypto = CryptoManager()
        self.__writer = Writer(self.__crypto, self.__StreamWriter, self.packetDict)

        LoginListPacket = Packet(
            0,
            0,
            "LOGINLIST",
            0,
            bson.encode(
                {
                    "appVer": "3.1.4",
                    "prtVer": "1",
                    "os": "win32",
                    "lang": "ko",
                    "duuid": self.device_uuid,
                    "oauthToken": self.__accessKey,
                    "dtype": 1,
                    "ntype": 0,
                    "MCCMNC": "999",
                    "revision": 0,
                    "chatIds": [],
                    "maxIds": [],
                    "lastTokenId": 0,
                    "lbk": 0,
                    "bg": False,
                }
            ),
        )

        if self.debug:
            print(bookingData)
            print(checkInData)
            print(LoginListPacket)

        self.__StreamWriter.write(self.__crypto.getHandshakePacket())

        self.loop.create_task(self.__writer.sendPacket(LoginListPacket))

        self.loop.create_task(self.__recvPacket())
        self.loop.create_task(self.__heartbeat())
        self.loop.create_task(self.onReady())

    def run(self):
        """
        A blocking function that logs in to the client and run.

        Parameters:

        Returns:

        Remarks:
        """
        self.loop = asyncio.get_event_loop()
        self.loop.create_task(self.__login(self.LoginId, self.LoginPw))
        self.loop.run_forever()


import json

import requests
import hashlib

Agent = "win32"
Lang = "ko"
Version = "3.1.1"
AppVersion = "3.1.1.2441"
OsVersion = "10.0"

AuthHeader = f"{Agent}/{Version}/{Lang}"
AuthUserAgent = f"KT/{Version} Wd/{OsVersion} {Lang}"

LoginUrl = "https://ac-sb-talk.kakao.com/win32/account/login.json"
RegisterDeviceUrl = "https://ac-sb-talk.kakao.com/win32/account/register_device.json"
RequestPasscodeUrl = "https://ac-sb-talk.kakao.com/win32/account/request_passcode.json"
MoreSettingUrl = f"https://sb-talk.kakao.com/win32/account/more_settings.json?since={0}&lang={Lang}"
MediaUrl = "https://up-m.talk.kakao.com/upload"


def RequestPasscode(email, password, device_name, device_uuid):
    h = Header(email, device_uuid)
    d = Data(email, password, device_name, device_uuid)
    d['permanent'] = "true"
    d['once'] = "false"
    r = requests.post(RequestPasscodeUrl, headers=h, data=d)

    return r.content.decode()


def RegisterDevice(email, password, device_name, device_uuid, passcode):
    h = Header(email, device_uuid)
    d = Data(email, password, device_name, device_uuid)
    d['permanent'] = "true"
    d['once'] = "false"
    d['passcode'] = passcode
    r = requests.post(RegisterDeviceUrl, headers=h, data=d)

    return r.content.decode()


def Login(email, password, device_name, device_uuid):
    h = Header(email, device_uuid)
    d = Data(email, password, device_name, device_uuid)
    d['permanent'] = True
    d['forced'] = True
    r = requests.post(LoginUrl, headers=h, data=d)

    return r.content.decode()


def Header(email, device_uuid):
    return {
        "Content-Type": "application/x-www-form-urlencoded",
        "A": AuthHeader,
        "X-VC": getXVC(email, device_uuid),
        "User-Agent": AuthUserAgent,
        "Accept": "*/*",
        "Accept-Language": Lang,
    }


def getXVC(email, device_uuid, isFull=False):
    string = f"HEATH|{AuthUserAgent}|DEMIAN|{email}|{device_uuid}".encode("utf-8")
    hash = hashlib.sha512(string).hexdigest()
    if isFull:
        return hash
    return hash[0:16]


def Data(email, password, device_name, device_uuid):
    return {
        "email": email,
        "password": password,
        "device_name": device_name,
        "device_uuid": device_uuid,
        "os_version": OsVersion
    }


def upload(data, dataType, userId):
    r = requests.post(MediaUrl,
                      headers={"A": AuthHeader, },
                      data={"attachment_type": dataType,
                            "user_id": userId, },
                      files={'attachment': data, })
    path = r.content.decode()
    key = path.replace('/talkm', "")
    url = "https://dn-m.talk.kakao.com"+path

    return path, key, url


def postText(chatId, li, text, notice, accessKey, deviceUUID):
    if li == 0:
        url = f"https://talkmoim-api.kakao.com/chats/{chatId}/posts"
    else:
        url = f"https://open.kakao.com/moim/chats/{chatId}/posts?link_id={li}"

    r = requests.post(url, headers={"A": AuthHeader,
                                    "User-Agent": AuthUserAgent,
                                    "Authorization": f"{accessKey}-{deviceUUID}",
                                    "Accept-Language": "ko"},
                      data={"content": json.dumps([{"text": text, "type": "text"}]),
                            "object_type": "TEXT", "notice": notice})

    print(r.content.decode())
